--[[
Interface: 1.5.1.0 b6732

Copyright (C) GtX (Andy), 2019

Author: GtX | Andy
Date: 05.10.2019

Contact:
https://forum.giants-software.com
https://github.com/GtX-Andy

History:
V 1.0.0.0 @ 05.10.2019 - FS19 Release Version

Note:
Idea based on FS15/FS17 PressureWasher.lua concept by Ifko[nator]

Important:
Free for use in other mods - no permission needed.
No changes are to be made to this script without permission from GtX | Andy

Frei verwendbar - keine erlaubnis nötig
An diesem Skript dürfen ohne Genehmigung von GtX | Andy keine Änderungen vorgenommen werden
]]


PressureWasherVehicle = {}

PressureWasherVehicleLance.BUILD_ID = 1
PressureWasherVehicleLance.VERSION_STRING = "1.0.0.0"

PressureWasherVehicle.MOD_NAME = g_currentModName

PressureWasherVehicle.STATE_OFF = 0
PressureWasherVehicle.STATE_MOVING = 1
PressureWasherVehicle.STATE_DEACTIVATE = 2
PressureWasherVehicle.STATE_ACTIVE = 3
PressureWasherVehicle.NUM_STATES = 4

PressureWasherVehicle.MAX_MOVE_DISTANCE = 1
PressureWasherVehicle.DEFAULT_WASH_MULTIPLIER = 1


function PressureWasherVehicle.initSpecialization()
    if PressureWasherVehicleLance == nil or PressureWasherVehicleStateEvent == nil or PressureWasherVehiclePumpEvent == nil then
        g_logManager:error("[%s]  Failed to load required files 'PressureWasherVehicleLance.lua' or 'PressureWasherVehicleStateEvent.lua' or 'PressureWasherVehiclePumpEvent.lua'!", PressureWasherVehicle.MOD_NAME)
    end

    if g_configurationManager:getConfigurationDescByName("pressureWasherLance") == nil then
        g_configurationManager:addConfigurationType("pressureWasherLance", g_i18n:getText("configuration_extension"), "pressureWasher", nil, nil, nil, ConfigurationUtil.SELECTOR_MULTIOPTION)
    end

    if g_configurationManager:getConfigurationDescByName("pressureWasherConsumer") == nil then
        g_configurationManager:addConfigurationType("pressureWasherConsumer", g_i18n:getText("configuration_consumer"), "pressureWasher", nil, nil, nil, ConfigurationUtil.SELECTOR_MULTIOPTION)
    end
end

function PressureWasherVehicle.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(FillUnit, specializations)
end

function PressureWasherVehicle.registerOverwrittenFunctions(vehicleType)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "loadDashboardGroupFromXML", PressureWasherVehicle.loadDashboardGroupFromXML)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "getIsInUse",                PressureWasherVehicle.getIsInUse)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "getIsDashboardGroupActive", PressureWasherVehicle.getIsDashboardGroupActive)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "getDrawFirstFillText",      PressureWasherVehicle.getDrawFirstFillText)
end

function PressureWasherVehicle.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onLoad",                     PressureWasherVehicle)
    SpecializationUtil.registerEventListener(vehicleType, "onDelete",                   PressureWasherVehicle)
    SpecializationUtil.registerEventListener(vehicleType, "onReadStream",               PressureWasherVehicle)
    SpecializationUtil.registerEventListener(vehicleType, "onWriteStream",              PressureWasherVehicle)
    SpecializationUtil.registerEventListener(vehicleType, "onUpdate",                   PressureWasherVehicle)
    SpecializationUtil.registerEventListener(vehicleType, "onFillUnitFillLevelChanged", PressureWasherVehicle)
end

function PressureWasherVehicle.registerFunctions(vehicleType)
    SpecializationUtil.registerFunction(vehicleType, "updatePressureWasherConsumers",           PressureWasherVehicle.updatePressureWasherConsumers)
    SpecializationUtil.registerFunction(vehicleType, "setPressureWasherState",                  PressureWasherVehicle.setPressureWasherState)
    SpecializationUtil.registerFunction(vehicleType, "setPressureWasherIsWashing",              PressureWasherVehicle.setPressureWasherIsWashing)
    SpecializationUtil.registerFunction(vehicleType, "setPressureWasherOperatingTime",          PressureWasherVehicle.setPressureWasherOperatingTime)
    SpecializationUtil.registerFunction(vehicleType, "adjustPumpWorkMultiplier",                PressureWasherVehicle.adjustPumpWorkMultiplier)
    SpecializationUtil.registerFunction(vehicleType, "getCanPressureWasherOperate",             PressureWasherVehicle.getCanPressureWasherOperate)
    SpecializationUtil.registerFunction(vehicleType, "getFormattedPressureWasherOperatingTime", PressureWasherVehicle.getFormattedPressureWasherOperatingTime)
    SpecializationUtil.registerFunction(vehicleType, "getPressureWasherOperatingTime",          PressureWasherVehicle.getPressureWasherOperatingTime)
    SpecializationUtil.registerFunction(vehicleType, "getIsPressureWasherActive",               PressureWasherVehicle.getIsPressureWasherActive)
    SpecializationUtil.registerFunction(vehicleType, "getIsPressureWasherUser",                 PressureWasherVehicle.getIsPressureWasherUser)
    SpecializationUtil.registerFunction(vehicleType, "getPumpWorkMultiplier",                   PressureWasherVehicle.getPumpWorkMultiplier)
    SpecializationUtil.registerFunction(vehicleType, "onPressureWasherUserDeleted",             PressureWasherVehicle.onPressureWasherUserDeleted)
end

function PressureWasherVehicle.getValidText(name, customEnv, rawText)
    if g_i18n:hasText(name, customEnv) then
        return g_i18n:getText(name, customEnv)
    end

    return rawText
end

function PressureWasherVehicle:onLoad(savegame)
    self.spec_pressureWasherVehicle = self[string.format("spec_%s.pressureWasherVehicle", PressureWasherVehicle.MOD_NAME)]

    if self.spec_pressureWasherVehicle == nil then
        g_logManager:error("[%s] Specialization with name 'pressureWasherVehicle' was not found in modDesc!", PressureWasherVehicle.MOD_NAME)
    end

    local spec = self.spec_pressureWasherVehicle
    local baseKey = "vehicle.pressureWasher"

    local lanceConfigurationId = Utils.getNoNil(self.configurations["pressureWasherLance"], 1)
    local lanceConfigurationKey = string.format("%s.pressureWasherLanceConfigurations.pressureWasherLanceConfiguration(%d)", baseKey, lanceConfigurationId - 1)

    ObjectChangeUtil.updateObjectChanges(self.xmlFile, baseKey .. ".pressureWasherLanceConfigurations.pressureWasherLanceConfiguration", lanceConfigurationId , self.components, self)

    -- Used to add/remove lance from vehicle.
    local interactionTrigger = I3DUtil.indexToObject(self.components, getXMLString(self.xmlFile, baseKey .. ".interactionTrigger#node"), self.i3dMappings)
    if interactionTrigger ~= nil then
        local colMask = getCollisionMask(interactionTrigger)
        if bitAND(Player.COLLISIONMASK_TRIGGER, colMask) == 0 then
            g_logManager:xmlWarning(self.configFilename, "Invalid collision mask for interaction trigger '%s.interactionTrigger#node'. Bit 20 needs to be set!", baseKey)
        end

        spec.activatable = PressureWasherVehicleActivatable:new(self, interactionTrigger)
    else
        g_logManager:xmlError(self.configFilename, "No interaction trigger found or given!")
    end

    -- XML filename for the hand tool. Must be using the <handToolType>pressureWasherVehicleLance</handToolType>.
    local handtoolFilename = getXMLString(self.xmlFile, baseKey .. ".pressureWasherLanceConfigurations#filename")
    if handtoolFilename ~= nil then
        self.handtoolFilename = Utils.getFilename(handtoolFilename, self.baseDirectory)

        if not fileExists(self.handtoolFilename) then
            g_logManager:xmlWarning(self.configFilename, "'handTool' %s could not be found at %s!", tostring(self.handtoolFilename))
            self.handtoolFilename = ""
        end
    else
        g_logManager:xmlWarning(self.configFilename, "No 'filename' given for %s.handtool", baseKey)
        self.handtoolFilename = ""
    end

    -- This node will be hidden when lance is in use.
    spec.lanceNode = I3DUtil.indexToObject(self.components, getXMLString(self.xmlFile, baseKey .. ".pressureWasherLanceConfigurations#node"), self.i3dMappings)

    -- This overwrites the 'washMultiplier' given in the handTool.xml only if > 0.
    spec.lanceWashMultiplier = Utils.getNoNil(getXMLFloat(self.xmlFile, lanceConfigurationKey .. "#washMultiplier"), 0)

    -- 15 is the standard for the placeable.
    spec.hoseLength = Utils.getNoNil(getXMLInt(self.xmlFile, lanceConfigurationKey .. "#hoseLength"), 15)

    spec.hoseLengthUsed = 0
    spec.lastInRangePosition = {0, 0, 0}

    -- This is the connection point to calculate the hose length used. May also add dynamic hose in future. The vehicle rootNode is used if nil.
    spec.hoseConnectionNode = Utils.getNoNil(I3DUtil.indexToObject(self.components, getXMLString(self.xmlFile, baseKey .. "#hoseConnectionNode"), self.i3dMappings), self.rootNode)

    spec.isActive = false
    spec.isWashing = false

    spec.activatePosition = nil

    spec.currentUser = nil
    spec.currentUserName = "N/A"

    spec.fillWarningText = g_i18n:getText("info_firstFillTheTool") .. " (%s)!"
    spec.inUseText = g_i18n:getText("shop_messageIsInUse") .. " (%s)!"

    spec.secureVehicleText = PressureWasherVehicle.getValidText("info_pwSecureVehicle", self.customEnvironment, "Secure vehicle from rolling away!")
    spec.hoseLengthText = PressureWasherVehicle.getValidText("info_pwHose", self.customEnvironment, "Hose")

    spec.pump = {}
    spec.pump.pressure = 0
    spec.pump.maxPressure = Utils.getNoNil(getXMLInt(self.xmlFile, baseKey .. "#maxPressure"), 4000)

    spec.pump.flowRate = 0
    spec.pump.maxFlowRate = 0

    -- Record the true operating time for the pressure washer.
    spec.pump.operatingTime = 0

    spec.pump.workMultiplier = 1
    spec.pump.canAdjustPressure = Utils.getNoNil(getXMLBool(self.xmlFile, baseKey .. "#canAdjustPressure"), false)

    -- Consumers are optional. Diesel and Water if used will be required for operation.
    -- All other fillTypes are treated as additives and will speed up wash times using 'washMultiplierIncrease' only if fill level > 0
    spec.consumers = {}
    spec.consumersByFillTypeIndex = {}
    spec.consumersByFillTypeName = {}

    local consumerConfigurationId = Utils.getNoNil(self.configurations["pressureWasherConsumer"], 1)
    local consumerConfigurationKey = string.format("%s.pressureWasherConsumerConfigurations.pressureWasherConsumerConfiguration(%d)", baseKey, consumerConfigurationId - 1)

    ObjectChangeUtil.updateObjectChanges(self.xmlFile, baseKey .. ".pressureWasherConsumerConfigurations.pressureWasherConsumerConfiguration", consumerConfigurationId , self.components, self)

    local i = 0
    while true do
        local key = string.format("%s.consumer(%d)", consumerConfigurationKey, i)
        if not hasXMLProperty(self.xmlFile, key) then
            break
        end

        local consumer = {isRequired = true, washMultiplierIncrease = 0, fillLevel = 0}

        consumer.fillUnitIndex = Utils.getNoNil(getXMLInt(self.xmlFile, key .. "#fillUnitIndex"), 1)

        local fillUnit = self:getFillUnitByIndex(consumer.fillUnitIndex)
        if fillUnit ~= nil then
            if fillUnit.capacity <= 0 then
                g_logManager:xmlWarning(self.configFilename, "FillUnit '%d' for consumer '%s capacity is 0!'", consumer.fillUnitIndex, key)
                break
            end
        else
            g_logManager:xmlWarning(self.configFilename, "Unknown fillUnit '%d' for consumer '%s'", consumer.fillUnitIndex, key)
            break
        end

        local fillTypeName = string.upper(Utils.getNoNil(getXMLString(self.xmlFile, key .. "#fillType"), "consumer"))
        local fillType = g_fillTypeManager:getFillTypeByName(fillTypeName)

        if fillType == nil then
            g_logManager:xmlWarning(self.configFilename, "Unknown fillType '%d' for consumer '%s'", fillTypeName, key)
            break
        end

        consumer.title = fillType.title
        consumer.fillTypeName = fillTypeName
        consumer.fillTypeIndex = fillType.index

        consumer.isDiesel = fillTypeName == "DIESEL"
        consumer.isWater = fillTypeName == "WATER"

        consumer.usagePerMS = math.max(Utils.getNoNil(getXMLFloat(self.xmlFile, key .. "#usagePerMin"), 15.0), 0.1) / (60 * 1000)

        if consumer.isWater then
            spec.pump.maxFlowRate = consumer.usagePerMS * 1000
        end

        if not consumer.isDiesel and not consumer.isWater then
            consumer.isRequired = false
            consumer.washMultiplierIncrease = math.max(Utils.getNoNil(getXMLFloat(self.xmlFile, key .. "#washMultiplierIncrease"), 0.1), 0.01)
        end

        table.insert(spec.consumers, consumer)

        spec.consumersByFillTypeName[fillTypeName] = consumer
        spec.consumersByFillTypeIndex[consumer.fillTypeIndex] = consumer

        i = i + 1
    end

    spec.numberOfConsumer = #spec.consumers
    spec.canDrawFirstFillText = Utils.getNoNil(getXMLBool(self.xmlFile, baseKey .. "#drawFirstFillText"), spec.numberOfConsumer > 0)

    if self.isClient then
        -- Animation nodes
        spec.animationNodes = g_animationManager:loadAnimations(self.xmlFile, baseKey .. ".animationNodes", self.components, self, self.i3dMappings)

        -- Sound samples
        spec.samples = {}
        spec.samples.start = g_soundManager:loadSampleFromXML(self.xmlFile, baseKey .. ".sounds", "start", self.baseDirectory, self.components, 1, AudioGroup.VEHICLE, self.i3dMappings, self)
        spec.samples.stop = g_soundManager:loadSampleFromXML(self.xmlFile, baseKey .. ".sounds", "stop", self.baseDirectory, self.components, 1, AudioGroup.VEHICLE, self.i3dMappings, self)
        spec.samples.idle = g_soundManager:loadSampleFromXML(self.xmlFile, baseKey .. ".sounds", "idle", self.baseDirectory, self.components, 0, AudioGroup.VEHICLE, self.i3dMappings, self)
        spec.samples.work = g_soundManager:loadSampleFromXML(self.xmlFile, baseKey .. ".sounds", "work", self.baseDirectory, self.components, 0, AudioGroup.VEHICLE, self.i3dMappings, self)

        -- Exhaust effect
        local linkNode = I3DUtil.indexToObject(self.components, getXMLString(self.xmlFile, baseKey .. ".exhaustEffect#node"), self.i3dMappings)
        if linkNode ~= nil then
            local filename = Utils.getNoNil(getXMLString(self.xmlFile, baseKey .. ".exhaustEffect#filename"), "$data/particleSystems/shared/exhaust.i3d")
            local i3dNode = g_i3DManager:loadSharedI3DFile(filename, self.baseDirectory, false, false, false)

            if i3dNode ~= 0 then
                local node = getChildAt(i3dNode, 0)

                if getHasShaderParameter(node, "param") then
                    spec.exhaustEffect = {effectNode = node, node = linkNode, filename = filename}

                    link(spec.exhaustEffect.node, spec.exhaustEffect.effectNode)
                    setVisibility(spec.exhaustEffect.effectNode, false)

                    local color = StringUtil.getVectorNFromString(Utils.getNoNil(getXMLString(self.xmlFile, baseKey .. ".exhaustEffect#color"), "0 0 0 1"), 4)
                    setShaderParameter(spec.exhaustEffect.effectNode, "exhaustColor", color[1], color[2], color[3], color[4], false)
                    spec.exhaustEffect.color = color

                    spec.exhaustEffect.scale = Utils.getNoNil(getXMLFloat(self.xmlFile, baseKey .. ".exhaustEffect#scale"), 1.0)
                    local xRot, zRot, _, _ = getShaderParameter(spec.exhaustEffect.effectNode, "param")
                    setShaderParameter(spec.exhaustEffect.effectNode, "param", xRot, zRot, 0, spec.exhaustEffect.scale, false)
                end

                delete(i3dNode)
            end
        end

        -- Turn on / off animation
        local turnOnAnimation = getXMLString(self.xmlFile, baseKey .. ".turnedAnimation#name")
        if turnOnAnimation ~= nil then
            if self:getAnimationExists(turnOnAnimation) then
                spec.turnOnAnimation = {}
                spec.turnOnAnimation.name = turnOnAnimation
                spec.turnOnAnimation.turnOnSpeedScale = Utils.getNoNil(getXMLFloat(self.xmlFile, baseKey .. ".turnedAnimation#turnOnSpeedScale"), 1)
                spec.turnOnAnimation.turnOffSpeedScale = Utils.getNoNil(getXMLFloat(self.xmlFile, baseKey .. ".turnedAnimation#turnOffSpeedScale"), -spec.turnOnAnimation.turnOnSpeedScale)
            else
                g_logManager:xmlWarning(self.configFileName, "Unable to find 'turnedAnimation' with name '%s'", turnOnAnimation)
            end
        end
    end

    local startDuration = 0
    if spec.samples ~= nil and spec.samples.start ~= nil then
        startDuration = spec.samples.start.duration
    end

    spec.startTime = 0
    spec.extendedStartTime = 0

    -- This allows the dashboard groups to have an extended time if required.
    spec.extendedStartTimeDuration = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.dashboard.groups#pressureWasherExtendedStartTime"), 2) * 1000

    -- Only required if a different time is required or no sound is used.
    spec.startDuration = Utils.getNoNil(getXMLFloat(self.xmlFile, baseKey .. ".startDuration"), startDuration)

    -- Dashboards (These will only update when pressure washer is active. It is best to have this invisible otherwise.)
    if self.loadDashboardsFromXML ~= nil then
        local dashboardsKey = "vehicle.pressureWasher.dashboards"

        self:loadDashboardsFromXML(self.xmlFile, dashboardsKey, {valueTypeToLoad = "time",
                                                                 valueObject = g_currentMission.environment,
                                                                 valueFunc = "getEnvironmentTime"})

        self:loadDashboardsFromXML(self.xmlFile, dashboardsKey, {valueTypeToLoad = "operatingTime",
                                                                 valueObject = self,
                                                                 valueFunc = "getFormattedPressureWasherOperatingTime"})

        self:loadDashboardsFromXML(self.xmlFile, dashboardsKey, {valueTypeToLoad = "waterPressure",
                                                                 valueObject = spec.pump,
                                                                 valueFunc = "pressure",
                                                                 minFunc = 0,
                                                                 maxFunc = "maxPressure"})

        if spec.numberOfConsumer > 0 then
            self:loadDashboardsFromXML(self.xmlFile, dashboardsKey, {valueTypeToLoad = "waterFlowRate",
                                                                     valueObject = spec.pump,
                                                                     valueFunc = "flowRate",
                                                                     minFunc = 0,
                                                                     maxFunc = "maxFlowRate"})
        end
    end

    if savegame ~= nil and not savegame.resetVehicles then
        local savegameKey = string.format("%s.%s.pressureWasherVehicle", savegame.key, PressureWasherVehicle.MOD_NAME)

        local operatingTime = Utils.getNoNil(getXMLFloat(savegame.xmlFile, savegameKey .. "#operatingTime"), spec.pump.operatingTime)
        self:setPressureWasherOperatingTime(operatingTime * 1000)

        local workMultiplier = getXMLFloat(savegame.xmlFile, savegameKey .. "#pumpWorkMultiplier")
        if workMultiplier ~= nil and workMultiplier > 0 then
            self:adjustPumpWorkMultiplier(workMultiplier, false)
        end
    end
end

function PressureWasherVehicle:onDelete()
    local spec = self.spec_pressureWasherVehicle

    self:setPressureWasherState(PressureWasherVehicle.STATE_OFF)

    if spec.activatable ~= nil then
        spec.activatable:delete()
        spec.activatable = nil
    end

    if self.isClient then
        if spec.animationNodes then
            g_animationManager:deleteAnimations(spec.animationNodes)
        end

        if spec.samples ~= nil then
            g_soundManager:deleteSamples(spec.samples)
        end

        if spec.exhaustEffect ~= nil then
            g_i3DManager:releaseSharedI3DFile(spec.exhaustEffect.filename, self.baseDirectory, true)
            spec.exhaustEffect = nil
        end
    end
end

function PressureWasherVehicle:onReadStream(streamId, connection)
    self:setPressureWasherOperatingTime(streamReadFloat32(streamId))
    self:adjustPumpWorkMultiplier(streamReadUInt8(streamId) / 10, false)

    if streamReadBool(streamId) then
        self:setPressureWasherState(PressureWasherVehicle.STATE_ACTIVE, nil, true)
    end
end

function PressureWasherVehicle:onWriteStream(streamId, connection)
    streamWriteFloat32(streamId, self:getPressureWasherOperatingTime())
    streamWriteUInt8(streamId, self:getPumpWorkMultiplier() * 10)

    -- No need to send 'currentUser' as players are loaded after vehicles anyway.
    streamWriteBool(streamId, self:getIsPressureWasherActive())
end

function PressureWasherVehicle:saveToXMLFile(xmlFile, key, usedModNames)
    setXMLFloat(xmlFile, key .. "#operatingTime", self:getPressureWasherOperatingTime() / 1000)
    setXMLFloat(xmlFile, key .. "#pumpWorkMultiplier", self:getPumpWorkMultiplier())
end

function PressureWasherVehicle:onUpdate(dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
    if self:getIsPressureWasherActive() then
        local spec = self.spec_pressureWasherVehicle

        if self.isServer then
            local movedDistance = 0

            if self.components[1] ~= nil and self.components[1].node ~= 0 then
                local x, y, z = getWorldTranslation(self.components[1].node)

                if spec.activatePosition == nil then
                    spec.activatePosition = {x, y, z}
                end

                movedDistance = MathUtil.vector3Length(spec.activatePosition[1] - x, spec.activatePosition[2] - y, spec.activatePosition[3] - z)
            end

            if movedDistance > PressureWasherVehicle.MAX_MOVE_DISTANCE then
                self:setPressureWasherState(PressureWasherVehicle.STATE_MOVING)
            end

            if VehicleDebug.state == VehicleDebug.DEBUG then
                g_currentMission:addExtraPrintText(string.format("Debug: Moved Distance: %.4f", movedDistance))
            end
        end

        if spec.currentUser ~= nil then
            local kx, _, kz = getWorldTranslation(spec.hoseConnectionNode)
            local px, py, pz = getWorldTranslation(spec.currentUser.rootNode)
            spec.hoseLengthUsed = MathUtil.vector2LengthSq(px - kx, pz - kz)

            -- Based on code from 'Placeable Pressure Washer' but Giants.
            if spec.hoseLengthUsed < (spec.hoseLength * spec.hoseLength) then
                spec.lastInRangePosition = {getTranslation(spec.currentUser.rootNode)}
            else
                local len = MathUtil.vector2Length(px - kx, pz - kz)
                local x, y, z = spec.lastInRangePosition[1],
                                spec.lastInRangePosition[2],
                                spec.lastInRangePosition[3]

                x = kx + ((px - kx) / len) * (spec.hoseLength - 0.00001 * dt)
                z = kz + ((pz - kz) / len) * (spec.hoseLength - 0.00001 * dt)

                spec.currentUser:moveToAbsoluteInternal(x, py, z)
                spec.lastInRangePosition = {x, y, z}

                if self:getIsPressureWasherUser() then
                    g_currentMission:showBlinkingWarning(g_i18n:getText("warning_hpwRangeRestriction"))
                end
            end
        end

        if spec.isWashing then
            spec.pump.flowRate = spec.pump.maxFlowRate * spec.pump.workMultiplier
            spec.pump.pressure = spec.pump.maxPressure * spec.pump.workMultiplier

            if self.isClient then
                if not spec.workSampleActive then
                    g_soundManager:playSample(spec.samples.work, 0)
                    spec.workSampleActive = true
                end
            end
        else
            spec.pump.flowRate = 0
            spec.pump.pressure = 0

            if self.isClient then
                if spec.workSampleActive then
                    g_soundManager:stopSample(spec.samples.work)
                    spec.workSampleActive = false
                end
            end
        end

        spec.pump.operatingTime = spec.pump.operatingTime + dt

        self:raiseActive()
    end
end

function PressureWasherVehicle:onFillUnitFillLevelChanged(fillUnitIndex, fillLevelDelta, fillTypeIndex, toolType, fillPositionData, appliedDelta)
    local spec = self.spec_pressureWasherVehicle
    local consumer = spec.consumersByFillTypeIndex[fillTypeIndex]

    if consumer ~= nil then
        -- Need to check each time fillLevel changes for the diesel consumer so also store fillLevel in spec for easy access.
        consumer.fillLevel = self:getFillUnitFillLevel(fillUnitIndex) or 0

        if consumer.fillLevel == 0 and consumer.isDiesel then
            if self:getIsPressureWasherUser() then
                self:setPressureWasherState(PressureWasherVehicle.STATE_OFF)
                g_currentMission:showBlinkingWarning(string.format(spec.fillWarningText, consumer.title), 3000)
            end
        end
    end
end

function PressureWasherVehicle:loadDashboardGroupFromXML(superFunc, xmlFile, key, group)
    if not superFunc(self, xmlFile, key, group) then
        return false
    end

    group.isPressureWasherRunning = Utils.getNoNil(getXMLBool(xmlFile, key .. "#isPressureWasherRunning"), false)
    group.isPressureWasherStarting = Utils.getNoNil(getXMLBool(xmlFile, key .. "#isPressureWasherStarting"), false)
    group.isPressureWasherExtendedStart = Utils.getNoNil(getXMLBool(xmlFile, key .. "#isPressureWasherExtendedStart"), false)

    return true
end

function PressureWasherVehicle:getIsInUse(superFunc, connection)
    local spec = self.spec_pressureWasherVehicle

    if spec.isActive and spec.ownerConnection ~= nil then
        if spec.ownerConnection ~= connection then
            return true
        end
    end

    return superFunc(self, connection)
end

function PressureWasherVehicle:getIsDashboardGroupActive(superFunc, group)
    local spec = self.spec_pressureWasherVehicle

    if group.isPressureWasherRunning and group.isPressureWasherStarting then
        if not spec.isActive then
            return false
        end
    end

    if group.isPressureWasherStarting and not group.isPressureWasherRunning then
        if not spec.isActive or spec.startTime < g_currentMission.time then
            return false
        end
    end

    if group.isPressureWasherRunning and not group.isPressureWasherStarting then
        if not spec.isActive or spec.startTime > g_currentMission.time then
            return false
        end
    end

    if group.isPressureWasherExtendedStart then
        if not spec.isActive or spec.extendedStartTime < g_currentMission.time then
            return false
        end
    end

    return superFunc(self, group)
end

function PressureWasherVehicle:getDrawFirstFillText(superFunc)
    local spec = self.spec_pressureWasherVehicle

    if self.isClient then
        if spec.canDrawFirstFillText and spec.numberOfConsumer > 0 then
            if self:getIsActiveForInput() and self:getIsSelected() then
                local consumer = spec.consumersByFillTypeName["DIESEL"]
                if consumer ~= nil and consumer.fillLevel <= 0 then
                    return true
                end

                consumer = spec.consumersByFillTypeName["WATER"]
                if consumer ~= nil and consumer.fillLevel <= 0 then
                    return true
                end
            end
        end
    end

    return false
end

function PressureWasherVehicle:updatePressureWasherConsumers(dt, isWashing, washMultiplier)
    washMultiplier = washMultiplier or PressureWasherVehicle.DEFAULT_WASH_MULTIPLIER

    if self.isServer then
        local spec = self.spec_pressureWasherVehicle

        if spec.numberOfConsumer > 0 then
            local pumpWorkMultiplier = spec.pump.workMultiplier
            local farmId = self:getOwnerFarmId()

            for _, consumer in ipairs (spec.consumers) do
                if consumer.isDiesel or isWashing then
                    local consumerUsage = consumer.usagePerMS  * dt

                    if consumer.isDiesel then
                        -- Update the 'fuelUsage' stats.
                        local farmStats = g_currentMission:farmStats(farmId)
                        farmStats:updateStats("fuelUsage", consumerUsage)
                    else
                        consumerUsage = consumerUsage * pumpWorkMultiplier

                        if consumer.washMultiplierIncrease > 0 and consumer.fillLevel > 0 then
                            washMultiplier = washMultiplier + consumer.washMultiplierIncrease
                        end
                    end

                    self:addFillUnitFillLevel(farmId, consumer.fillUnitIndex, -consumerUsage, consumer.fillTypeIndex, ToolType.UNDEFINED, nil)
                end
            end

            return washMultiplier * pumpWorkMultiplier
        end
    end

    return washMultiplier
end

function PressureWasherVehicle:setPressureWasherState(state, currentUser, noEventSend)
    local spec = self.spec_pressureWasherVehicle
    local isActive = false

    if state == PressureWasherVehicle.STATE_ACTIVE then
        isActive = true
    elseif state == PressureWasherVehicle.STATE_MOVING then
        if self:getIsPressureWasherUser() then
            g_currentMission:showBlinkingWarning(spec.secureVehicleText, 3000)
        end
    end

    if spec.isActive ~= isActive then
        PressureWasherVehicleStateEvent.sendEvent(self, state, currentUser, noEventSend)

        spec.isActive = isActive

        if spec.isActive then
            if currentUser ~= nil then
                spec.currentUser = currentUser
                spec.currentUserName = currentUser.visualInformation.playerName

                -- No need to sync this as I have just sent the 'currentUser' to all players anyway.
                spec.currentUser:equipHandtool(spec.handtoolFilename, true, true)

                local currentHandtool = currentUser.baseInformation.currentHandtool
                if currentHandtool ~= nil and currentHandtool.setConnectedVehicle ~= nil then
                    currentHandtool:setConnectedVehicle(self, spec.lanceWashMultiplier, spec.pump.canAdjustPressure)
                end

                if self.isServer then
                    spec.ownerConnection = currentUser.networkInformation.creatorConnection
                    spec.currentUser:addDeleteListener(self, "onPressureWasherUserDeleted")
                end
            end

            spec.startTime = g_currentMission.time + spec.startDuration
            spec.extendedStartTime = spec.startTime + spec.extendedStartTimeDuration
        else
            if spec.currentUser ~= nil then
                -- Not required on 'STATE_DEACTIVATE' as tool is already being unequipped and this should already be synced.
                if state == PressureWasherVehicle.STATE_OFF or state == PressureWasherVehicle.STATE_MOVING then
                    spec.currentUser:equipHandtool("", true, noEventSend)
                end

                if self.isServer then
                    spec.ownerConnection = nil
                    spec.currentUser:removeDeleteListener(self, "onPressureWasherUserDeleted")
                end
            end

            spec.currentUser = nil
            spec.currentUserName = "N/A"

            spec.isWashing = false

            spec.pump.flowRate = 0
            spec.pump.pressure = 0

            spec.activatePosition = nil
        end

        if self.isClient then
            if spec.lanceNode ~= nil then
                setVisibility(spec.lanceNode, not spec.isActive)
            end

            if spec.samples ~= nil then
                if spec.isActive then
                    g_soundManager:stopSample(spec.samples.stop)
                    g_soundManager:playSample(spec.samples.start)
                    g_soundManager:playSample(spec.samples.idle, 0, spec.samples.start)
                else
                    g_soundManager:stopSample(spec.samples.start)
                    g_soundManager:stopSample(spec.samples.idle)

                    if spec.workSampleActive then
                        g_soundManager:stopSample(spec.samples.work)
                        spec.workSampleActive = false
                    end

                    g_soundManager:playSample(spec.samples.stop)
                end
            end

            if self.playAnimation ~= nil and spec.turnOnAnimation ~= nil then
                if spec.isActive then
                    self:playAnimation(spec.turnOnAnimation.name, spec.turnOnAnimation.turnOnSpeedScale, self:getAnimationTime(spec.turnOnAnimation.name), true)
                else
                    self:playAnimation(spec.turnOnAnimation.name, spec.turnOnAnimation.turnOffSpeedScale, self:getAnimationTime(spec.turnOnAnimation.name), true)
                end
            end

            if spec.animationNodes ~= nil then
                if spec.isActive then
                    g_animationManager:startAnimations(spec.animationNodes)
                else
                    g_animationManager:stopAnimations(spec.animationNodes)
                end
            end

            if spec.exhaustEffect ~= nil then
                setVisibility(spec.exhaustEffect.effectNode, spec.isActive)
            end
        end

        self:raiseActive()
    end
end

function PressureWasherVehicle:setPressureWasherIsWashing(isWashing)
    self.spec_pressureWasherVehicle.isWashing = Utils.getNoNil(isWashing, false)
end

function PressureWasherVehicle:setPressureWasherOperatingTime(operatingTime, increase)
    if operatingTime ~= nil then
        self.spec_pressureWasherVehicle.pump.operatingTime = math.max(operatingTime, 0)
    end
end

function PressureWasherVehicle:adjustPumpWorkMultiplier(value, isInputValue)
    local spec = self.spec_pressureWasherVehicle

    if value ~= nil and value ~= 0 then
        if isInputValue then
            if spec.pump.canAdjustPressure then
                local workMultiplier = spec.pump.workMultiplier

                if value > 0 then
                    workMultiplier = math.min(workMultiplier + 0.1, 1.0)
                else
                    workMultiplier = math.max(workMultiplier - 0.1, 0.1)
                end

                if workMultiplier ~= spec.pump.workMultiplier then
                    spec.pump.workMultiplier = workMultiplier

                    if g_server ~= nil then
                        g_server:broadcastEvent(PressureWasherVehiclePumpEvent:new(self, workMultiplier), nil, nil, self)
                    else
                        g_client:getServerConnection():sendEvent(PressureWasherVehiclePumpEvent:new(self, workMultiplier))
                    end
                end
            end
        else
            spec.pump.workMultiplier = MathUtil.clamp(MathUtil.round(value, 1), 0.1, 1.0)
        end
    end

    return spec.pump.workMultiplier
end

function PressureWasherVehicle:getCanPressureWasherOperate(activatePressed, showWarning)
    local spec = self.spec_pressureWasherVehicle

    if activatePressed and spec.startTime < g_currentMission.time then
        for _, consumer in ipairs (spec.consumers) do
            if consumer.isRequired and consumer.fillLevel <= 0 then
                if showWarning and self:getIsPressureWasherUser() then
                    g_currentMission:showBlinkingWarning(string.format(spec.fillWarningText, consumer.title), 3000)
                end

                return false
            end
        end
    else
        return false
    end

    return true
end

function PressureWasherVehicle:getFormattedPressureWasherOperatingTime()
    local minutes = self:getPressureWasherOperatingTime() / (1000 * 60)
    local hours = math.floor(minutes / 60)

    minutes = math.floor((minutes - hours * 60) / 6)
    local minutesString = string.format("%02d", minutes * 10)

    return tonumber(hours .. "." .. minutesString)
end

function PressureWasherVehicle:getPressureWasherOperatingTime()
    return self.spec_pressureWasherVehicle.pump.operatingTime or 0
end

function PressureWasherVehicle:getIsPressureWasherActive()
    return self.spec_pressureWasherVehicle.isActive
end

function PressureWasherVehicle:getIsPressureWasherUser()
    return self.spec_pressureWasherVehicle.currentUser == g_currentMission.player
end

function PressureWasherVehicle:getPumpWorkMultiplier()
    return self.spec_pressureWasherVehicle.pump.workMultiplier or 1
end

function PressureWasherVehicle:onPressureWasherUserDeleted(player)
    local spec = self.spec_pressureWasherVehicle

    if spec.isActive and spec.currentUser == player then
        self:setPressureWasherState(PressureWasherVehicle.STATE_OFF)
    end
end


PressureWasherVehicleActivatable = {}
local PressureWasherVehicleActivatable_mt = Class(PressureWasherVehicleActivatable)

function PressureWasherVehicleActivatable:new(vehicle, interactionTrigger)
    local self = {}
    setmetatable(self, PressureWasherVehicleActivatable_mt)

    self.vehicle = vehicle

    self.interactionTrigger = interactionTrigger
    addTrigger(self.interactionTrigger, "interactionTriggerCallback", self)

    self.isEnabled = true
    self.activated = false

    self.activateText = g_i18n:getText("typeDesc_highPressureWasher")

    self.turnOnText = string.format(g_i18n:getText("action_turnOnOBJECT"), self.activateText)
    self.turnOffText = string.format(g_i18n:getText("action_turnOffOBJECT"), self.activateText)

    self.detailsText = g_i18n:getText("ui_modHubDetails")

    return self
end

function PressureWasherVehicleActivatable:delete()
    self.isEnabled = false

    if self.interactionTrigger ~= nil then
        removeTrigger(self.interactionTrigger)
        self.interactionTrigger = nil
    end

    if self.activated then
        g_currentMission:removeActivatableObject(self)
        self.activated = false
    end
end

function PressureWasherVehicleActivatable:getIsActivatable()
    if self.isEnabled and g_currentMission.controlPlayer then
        if g_currentMission.accessHandler:canFarmAccess(g_currentMission:getFarmId(), self.vehicle) then
            self:updateActivateText()

            return true
        end
    end

    return false
end

function PressureWasherVehicleActivatable:onActivateObject()
    local spec = self.vehicle.spec_pressureWasherVehicle

    if spec.isActive then
        if self.vehicle:getIsPressureWasherUser() then
            self.vehicle:setPressureWasherState(PressureWasherVehicle.STATE_OFF)
        else
            g_currentMission:showBlinkingWarning(string.format(spec.inUseText, spec.currentUserName), 3000)
        end
    else
        -- Remove any hand tool first in case the have listeners that will override me. Extra data but worth it.
        g_currentMission.player:unequipHandtool()

        -- Now try and start vehicle and equip lance.
        local dieselConsumer = spec.consumersByFillTypeName["DIESEL"]
        if dieselConsumer == nil or dieselConsumer.fillLevel > 0 then
            self.vehicle:setPressureWasherState(PressureWasherVehicle.STATE_ACTIVE, g_currentMission.player)
        else
            g_currentMission:showBlinkingWarning(string.format(spec.fillWarningText, dieselConsumer.title), 3000)
        end
    end
end

function PressureWasherVehicleActivatable:drawActivate()
    return
end

function PressureWasherVehicleActivatable:updateActivateText()
    if self.vehicle:getIsPressureWasherActive() then
        if self.vehicle:getIsPressureWasherUser() then
            self.activateText = self.turnOffText
        else
            self.activateText = self.detailsText
        end
    else
        self.activateText = self.turnOnText
    end
end

function PressureWasherVehicleActivatable:shouldRemoveActivatable()
    return false
end

function PressureWasherVehicleActivatable:interactionTriggerCallback(triggerId, otherId, onEnter, onLeave, onStay)
    if self.isEnabled and (onEnter or onLeave) then
        if g_currentMission.player ~= nil and otherId == g_currentMission.player.rootNode then
            if onEnter then
                if not self.activated and g_currentMission.accessHandler:canFarmAccess(g_currentMission:getFarmId(), self.vehicle) then
                    g_currentMission:addActivatableObject(self)
                    self.activated = true
                end
            else
                if self.activated then
                    g_currentMission:removeActivatableObject(self)
                    self.activated = false
                end
            end
        end
    end
end
