--[[
Interface: 1.5.1.0 b6732

Author: GtX | Andy
Date: 06.10.2019
Version: 1.0.0.0

Contact:
https://forum.giants-software.com
https://github.com/GtX-Andy

History:
V 1.0.0.0 @ 06.10.2019 - FS19 Release Version

About:
Based on 'HighPressureWasherLance.lua' by Giants Software, not using the class in case another mod modifies it and I have some extra animations.
https://gdn.giants-software.com/documentation_scripting_fs19.php?version=script&category=63&class=10272

Important:
No changes are to be made to this script without permission from GtX | Andy
An diesem Skript dürfen ohne Genehmigung von GtX | Andy keine Änderungen vorgenommen werden
]]


PressureWasherVehicleLance = {}

PressureWasherVehicleLance.BUILD_ID = 1
PressureWasherVehicleLance.VERSION_STRING = "1.0.0.0"

PressureWasherVehicleLance.MOD_NAME = g_currentModName
PressureWasherVehicleLance.COLLISION_MASK = 32 + 64 + 128 + 256 + 4096 + 8194

local PressureWasherVehicleLance_mt = Class(PressureWasherVehicleLance, HandTool)
InitObjectClass(PressureWasherVehicleLance, "PressureWasherVehicleLance")

function PressureWasherVehicleLance:new(isServer, isClient, customMt)
    local self = PressureWasherVehicleLance:superClass().new(self, isServer, isClient, customMt or PressureWasherVehicleLance_mt)

    self.isPWVLance = true

    self.connectedVehicle = nil
    self.distanceToVehicle = 0

    self.allowPressureAdjustment = false

    return self
end

function PressureWasherVehicleLance:load(xmlFilename, player)
    if not PressureWasherVehicleLance:superClass().load(self, xmlFilename, player) then
        return false
    end

    local xmlFile = loadXMLFile("TempXML", xmlFilename)

    self.lanceNode = I3DUtil.indexToObject(self.rootNode, getXMLString(xmlFile, "handTool.pressureWasherVehicleLance.lance#node"))
    self.lanceRaycastNode = I3DUtil.indexToObject(self.rootNode, getXMLString(xmlFile, "handTool.pressureWasherVehicleLance.lance#raycastNode"))

    self.washDistance = Utils.getNoNil(getXMLFloat(xmlFile, "handTool.pressureWasherVehicleLance.lance#washDistance"), 10)
    self.washMultiplier = Utils.getNoNil(getXMLFloat(xmlFile, "handTool.pressureWasherVehicleLance.lance#washMultiplier"), 1)

    self.effects = g_effectManager:loadEffect(xmlFile, "handTool.pressureWasherVehicleLance.effects", self.rootNode, self)
    g_effectManager:setFillType(self.effects, FillType.WATER)

    self.washingSample = g_soundManager:loadSampleFromXML(xmlFile, "handTool.pressureWasherVehicleLance.sounds", "washing", self.baseDirectory, self.rootNode, 0, AudioGroup.VEHICLE, nil, self)

    -- Allow trigger / handle movement.
    local node = I3DUtil.indexToObject(self.rootNode, getXMLString(xmlFile, "handTool.pressureWasherVehicleLance.lance.handle#node"))
    if node ~= nil then
        self.handle = {}
        self.handle.node = node

        local rotOff = StringUtil.getRadiansFromString(getXMLString(xmlFile, "handTool.pressureWasherVehicleLance.lance.handle#rotOff"), 3)
        self.handle.rotOff = Utils.getNoNil(rotOff, {getRotation(node)})
        self.handle.rotOn = Utils.getNoNil(StringUtil.getRadiansFromString(getXMLString(xmlFile, "handTool.pressureWasherVehicleLance.lance.handle#rotOn"), 3), self.handle.rotOff)

        setRotation(self.handle.node, self.handle.rotOff[1], self.handle.rotOff[2], self.handle.rotOff[3])
    end

    self.loadPressureInputBindings = Utils.getNoNil(getXMLBool(xmlFile, "handTool.pressureWasherVehicleLance#loadPressureInputBindings"), false)

    -- Allow pressure knob movement and effect scaling.
    if self.loadPressureInputBindings then
        node = I3DUtil.indexToObject(self.rootNode, getXMLString(xmlFile, "handTool.pressureWasherVehicleLance.lance.pressureAdjustment#rotateNode"))
        if node ~= nil then
            self.pressureAdjustment = {}
            self.pressureAdjustment.node = node

            local increments = getXMLString(xmlFile, "handTool.pressureWasherVehicleLance.lance.pressureAdjustment#rotateIncrements")
            self.pressureAdjustment.increments = StringUtil.getRadiansFromString(Utils.getNoNil(increments, "0 0 -10"), 3)
            self.pressureAdjustment.startRotation = {getRotation(node)}

            -- Only with 'rotateNode'. Default scale '1 1 1' used only as the effect node should not be scaled in GE if mod is correctly made.
            self.pressureAdjustment.effectNode = I3DUtil.indexToObject(self.rootNode, getXMLString(xmlFile, "handTool.pressureWasherVehicleLance.lance.pressureAdjustment#effectNode"))
        end
    end

    delete(xmlFile)

    return true
end

function PressureWasherVehicleLance:delete()
    g_effectManager:deleteEffects(self.effects)
    g_soundManager:deleteSample(self.washingSample)

    -- Try and call again in case 'onDeactivate' is not called by a faulty mod.
    if self:hasConnectedVehicle() then
        self.connectedVehicle:setPressureWasherState(PressureWasherVehicle.STATE_DEACTIVATE)
        self.connectedVehicle = nil
    end

    PressureWasherVehicleLance:superClass().delete(self)
end

function PressureWasherVehicleLance:onDeactivate()
    PressureWasherVehicleLance:superClass().onDeactivate(self)

    self:setIsWashing(false, true, true)

    if self:hasConnectedVehicle() then
        self.connectedVehicle:setPressureWasherState(PressureWasherVehicle.STATE_DEACTIVATE)
        self.connectedVehicle = nil
    end
end

function PressureWasherVehicleLance:draw()
    PressureWasherVehicleLance:superClass().draw(self)

    if self:getIsActiveForInput() and self:hasConnectedVehicle() then
        local spec = self.connectedVehicle.spec_pressureWasherVehicle

        for _, consumer in ipairs (spec.consumers) do
            local fillUnit = self.connectedVehicle.spec_fillUnit.fillUnits[consumer.fillUnitIndex]
            if fillUnit ~= nil then
                g_currentMission:addExtraPrintText(string.format("%s: %d (%d%%)", consumer.title, MathUtil.round(fillUnit.fillLevel), 100 * (fillUnit.fillLevel / fillUnit.capacity)))
            end
        end

        if self.allowPressureAdjustment then
            local setWorkingPressure = MathUtil.round(spec.pump.maxPressure * spec.pump.workMultiplier)
            g_currentMission:addExtraPrintText(string.format("PSI: %d (%d%%)", setWorkingPressure, 100 * (setWorkingPressure / spec.pump.maxPressure)))
        end

        -- Show the percentage of the hose that is used.
        g_currentMission:addExtraPrintText(string.format("%s: %.0f%%", spec.hoseLengthText, 100 * math.min(math.sqrt(spec.hoseLengthUsed) / spec.hoseLength), 1))
    end
end

function PressureWasherVehicleLance:update(dt, allowInput)
    PressureWasherVehicleLance:superClass().update(self, dt, allowInput)

    if self:hasConnectedVehicle() then
        if allowInput then
            self:setIsWashing(self.connectedVehicle:getCanPressureWasherOperate(self.activatePressed, true), false)
        end

        if self.isServer then
            local washMultiplier = self.connectedVehicle:updatePressureWasherConsumers(dt, self.doWashing, self.washMultiplier)

            if self.doWashing then
                self.foundVehicle = nil
                self:cleanVehicle(self.player.cameraNode, dt, washMultiplier)

                if self.lanceRaycastNode ~= nil then
                    self:cleanVehicle(self.lanceRaycastNode, dt, washMultiplier)
                end
            end

            if self:getIsDebugActive() then
                if self:getIsActiveForInput() then
                    local vehicleName = "N/A"
                    local distanceToVehicle = "N/A"
                    local spec = self.connectedVehicle.spec_pressureWasherVehicle

                    if self.doWashing and self.foundVehicle ~= nil then
                        vehicleName = self.foundVehicle:getName()
                        distanceToVehicle = self.distanceToVehicle
                    end

                    local values = {
                        {name = "Pressure Washer Name", value = self.connectedVehicle:getName()},
                        {name = "Pump Operating Time", value = self.connectedVehicle:getFormattedPressureWasherOperatingTime()},
                        {name = "Wash Multiplier", value = self.washMultiplier},
                        {name = "Extra Wash Multiplier", value = washMultiplier - self.washMultiplier},
                        {name = "Total Wash Multiplier", value = washMultiplier},
                        {name = "Hose Length", value = spec.hoseLength},
                        {name = "Hose Length Used", value = MathUtil.round(spec.hoseLengthUsed)},
                        {name = "Vehicle Name", value = vehicleName},
                        {name = "Distance To Vehicle", value = distanceToVehicle}
                    }

                    if self.doWashing and self.foundVehicle ~= nil then
                        spec = self.foundVehicle.spec_washable

                        if spec ~= nil and spec.washableNodes ~= nil then
                            for i, nodeData in ipairs(spec.washableNodes) do
                                local changedAmount = nodeData.updateFunc(self.foundVehicle, nodeData, 3600000)
                                table.insert(values, {name = "WashableNode" .. i, value = string.format("%.4f a/h (%.2f)", changedAmount, self.foundVehicle:getNodeDirtAmount(nodeData))})
                            end
                        end
                    end

                    DebugUtil.renderTable(0.9, 0.55, 0.02, values, 0)
                end
            end
        end

        if self.pressureAdjustment ~= nil then
            local pumpWorkMultiplier = self.connectedVehicle:getPumpWorkMultiplier()

            if pumpWorkMultiplier ~= self.pumpWorkMultiplier then
                self.pumpWorkMultiplier = pumpWorkMultiplier

                local factor = 10 * (pumpWorkMultiplier - 0.1)
                local iX, iY, iZ = unpack(self.pressureAdjustment.increments)
                local rX, rY, rZ = unpack(self.pressureAdjustment.startRotation)

                setRotation(self.pressureAdjustment.node, rX + (iX * factor), rY + (iY * factor), rZ + (iZ * factor))

                if self.pressureAdjustment.effectNode ~= nil then
                    if pumpWorkMultiplier < 1 then
                        pumpWorkMultiplier = pumpWorkMultiplier + 0.08
                    end

                    setScale(self.pressureAdjustment.effectNode, pumpWorkMultiplier, pumpWorkMultiplier, pumpWorkMultiplier)
                end
            end
        end
    end

    self.activatePressed = false
end

function PressureWasherVehicleLance:setIsWashing(doWashing, force, noEventSend)
    HPWLanceStateEvent.sendEvent(self.player, doWashing, noEventSend)

    if self.doWashing ~= doWashing then
        if doWashing then
            g_effectManager:setFillType(self.effects, FillType.WATER)
            g_effectManager:startEffects(self.effects)

            g_soundManager:playSample(self.washingSample)

            if self.handle ~= nil then
                setRotation(self.handle.node, self.handle.rotOn[1], self.handle.rotOn[2], self.handle.rotOn[3])
            end
        else
            if force then
                g_effectManager:resetEffects(self.effects)
            else
                g_effectManager:stopEffects(self.effects)
            end

            g_soundManager:stopSample(self.washingSample)

            if self.handle ~= nil then
                setRotation(self.handle.node, self.handle.rotOff[1], self.handle.rotOff[2], self.handle.rotOff[3])
            end
        end

        self.doWashing = doWashing

        if self:hasConnectedVehicle() then
            self.connectedVehicle:setPressureWasherIsWashing(doWashing)
        end
    end
end

function PressureWasherVehicleLance:cleanVehicle(node, dt, washMultiplier)
    local x, y, z = getWorldTranslation(node)
    local dx, dy, dz = localDirectionToWorld(node, 0, 0, -1)
    local lastFoundVehicle = self.foundVehicle

    raycastAll(x, y, z, dx, dy, dz, "washRaycastCallback", self.washDistance, self, PressureWasherVehicleLance.COLLISION_MASK)

    if self.foundVehicle ~= nil and lastFoundVehicle ~= self.foundVehicle then
        self.foundVehicle:addDirtAmount(-washMultiplier * dt / self.foundVehicle:getWashDuration())
    end
end

function PressureWasherVehicleLance:washRaycastCallback(hitActorId, x, y, z, distance, nx, ny, nz, subShapeIndex, hitShapeId)
    local vehicle = g_currentMission.nodeToObject[hitActorId]

    if hitActorId ~= hitShapeId then
        -- object is a compoundChild. Try to find the compound
        local parentId = hitShapeId
        while parentId ~= 0 do
            if g_currentMission.nodeToObject[parentId] ~= nil then
                -- found valid compound
                vehicle = g_currentMission.nodeToObject[parentId]
                break
            end

            parentId = getParent(parentId)
        end
    end

    self.distanceToVehicle = distance

    if vehicle ~= nil and vehicle.getAllowsWashingByType ~= nil and vehicle:getAllowsWashingByType(Washable.WASHTYPE_HIGH_PRESSURE_WASHER) then
        self.foundVehicle = vehicle

        return false
    end

    return true
end

function PressureWasherVehicleLance:getIsActiveForInput()
    if self.player == g_currentMission.player and not g_gui:getIsGuiVisible() then
        return true
    end

    return false
end

function PressureWasherVehicleLance:setConnectedVehicle(vehicle, washMultiplier, allowPressureAdjustment)
    self.connectedVehicle = nil

    if vehicle ~= nil and vehicle.spec_pressureWasherVehicle ~= nil then
        if washMultiplier ~= nil and washMultiplier > 0 then
            self.washMultiplier = washMultiplier
        end

        self.allowPressureAdjustment = Utils.getNoNil(allowPressureAdjustment, false)

        if not self.allowPressureAdjustment and self.loadPressureInputBindings then
            if self.actionEventIdIncreaseWorkingPressure ~= nil then
                g_inputBinding:setActionEventTextVisibility(self.actionEventIdIncreaseWorkingPressure, false)
                g_inputBinding:setActionEventActive(self.actionEventIdIncreaseWorkingPressure, false)
            end

            if self.actionEventIdDecreaseWorkingPressure ~= nil then
                g_inputBinding:setActionEventTextVisibility(self.actionEventIdDecreaseWorkingPressure, false)
                g_inputBinding:setActionEventActive(self.actionEventIdDecreaseWorkingPressure, false)
            end
        end

        self.connectedVehicle = vehicle

        self.isActive = true
        self:raiseActive()
    else
        if self:getIsDebugActive() then
            print(string.format("  Debug: [%s]  Failed to set vehicle or vehicle removed!", PressureWasherVehicleLance.MOD_NAME))
        end
    end
end

function PressureWasherVehicleLance:getConnectedVehicle()
    if self.connectedVehicle ~= nil and not self.connectedVehicle.isDeleted then
        return self.connectedVehicle
    end

    return nil
end

function PressureWasherVehicleLance:hasConnectedVehicle()
    if self.connectedVehicle ~= nil then
        return not self.connectedVehicle.isDeleted
    end

    return false
end

function PressureWasherVehicleLance:registerActionEvents()
    if self.loadPressureInputBindings then
        g_inputBinding:beginActionEventsModification(Player.INPUT_CONTEXT_NAME)

        local _, actionEventId = g_inputBinding:registerActionEvent(InputAction.TOGGLE_TURNLIGHT_RIGHT, self, self.onIncreaseWorkingPressure, false, true, false, true, 1)
        g_inputBinding:setActionEventText(actionEventId, "PSI (+)")
        self.actionEventIdIncreaseWorkingPressure = actionEventId

        _, actionEventId = g_inputBinding:registerActionEvent(InputAction.TOGGLE_TURNLIGHT_LEFT, self, self.onDecreaseWorkingPressure, false, true, false, true, -1)
        g_inputBinding:setActionEventText(actionEventId, "PSI (-)")
        self.actionEventIdDecreaseWorkingPressure = actionEventId

        g_inputBinding:endActionEventsModification()
    end
end

function PressureWasherVehicleLance:onIncreaseWorkingPressure(actionName, inputValue, callbackState, isAnalog)
    if self:hasConnectedVehicle() then
        self.connectedVehicle:adjustPumpWorkMultiplier(0.1, true)
    end
end

function PressureWasherVehicleLance:onDecreaseWorkingPressure(actionName, inputValue, callbackState, isAnalog)
    if self:hasConnectedVehicle() then
        self.connectedVehicle:adjustPumpWorkMultiplier(-0.1, true)
    end
end

function PressureWasherVehicleLance:isBeingUsed()
    return self.doWashing
end

function PressureWasherVehicleLance:getIsDebugActive()
    return VehicleDebug.state == VehicleDebug.DEBUG
end

registerHandTool(string.format("%s.pressureWasherVehicleLance", PressureWasherVehicleLance.MOD_NAME), PressureWasherVehicleLance)
